module Main exposing (..)

import Browser
import Html exposing (..)
import Html.Attributes exposing (..)
import Html.Events exposing (..)
import Svg exposing (svg)
import Svg.Attributes
import Character exposing (Character)
import Stats exposing (Stat)
import BoundedInt exposing (BoundedInt)
import Stats exposing (Category)
import Creation exposing (CreationPoints)
import OrderedCategories as OCat exposing (OCat, CategoriesType)
import Http
import Json.Decode
import Url
import Browser.Navigation as Nav
import Dice exposing (Dice)
import Random

main : Program () Model Msg
main = 
    Browser.application
        { init = init
        , view = view
        , update = update
        , subscriptions = subscriptions
        , onUrlChange = UrlChanged
        , onUrlRequest = LinkClicked
        }

init : () -> Url.Url -> Nav.Key -> ( Model, Cmd Msg )
init _ url key =
    (   { character = Character.new
        , modal = Nothing
        , modalValue = ""
        , activeDropdown = Nothing
        , key = key
        , url = url
        , dice = Dice.new
        , roll = []
        }
    , Cmd.none
    )

type alias Model = 
    { character : Character
    , modal : Maybe ModalType
    , modalValue : String
    , activeDropdown : Maybe ActiveDropdown
    , key : Nav.Key
    , url : Url.Url
    , dice : Dice
    , roll : List Int
    }

type ActiveDropdown
    = Backgrounds

type Mode
    = CreateCharacter
    | ViewCharacter
    | RollDice

type Msg 
    = OpenModal ModalType
    | CloseModal
    | ChangeModalValue String
    | SaveModalChange ModalType
    | ChangeStat Stat
    | ChangePriority CategoriesType Category Int
    | SetDropDownActive ( Maybe ActiveDropdown )
    | SetAffinity Stat
    | ImportCharacter String
    | JsonContent ( Result Http.Error Character )
    | LinkClicked Browser.UrlRequest
    | UrlChanged Url.Url
    | ResetCharacter
    | SetDice Dice
    | Roll
    | RollResult ( List Int )
    | ResetRoll

type ModalType
    = Name

update : Msg -> Model -> ( Model, Cmd Msg )
update msg model = 
    case msg of
        OpenModal modalType -> 
            (   { model 
                | modal = ( Just modalType ) 
                , modalValue = modalValue modalType model.character
                , activeDropdown = Nothing
                }
            , Cmd.none
            )
        CloseModal ->
            ( { model | modal = Nothing }
            , Cmd.none
            )
        ChangeModalValue value ->
            ( { model | modalValue = value }
            , Cmd.none
            )
        SaveModalChange modalType ->
            case modalType of
                Name ->
                    (   { model 
                        | character = Character.changeName model.character model.modalValue 
                        , modal = Nothing
                        }
                    , Cmd.none
                    )
        ChangeStat stat ->
            (   { model 
                | character = Character.changeStat model.character stat
                , activeDropdown = Nothing
                }
            , Cmd.none
            )
        ChangePriority catType category value ->
            (   { model 
                | character = Character.changeOrder model.character catType category value
                , activeDropdown = Nothing
                }
            , Cmd.none
            )
        SetDropDownActive value ->
            (   { model
                | activeDropdown = value
                }
            , Cmd.none
            )
        SetAffinity stat ->
            (   { model
                | character = Character.setAffinity model.character stat
                }
            , Cmd.none
            )
        ImportCharacter url ->
            (   { model
                | activeDropdown = Nothing
                }
            , Http.get
                { url = url
                , expect = Http.expectJson JsonContent jsonDecoder
                }
            )
        JsonContent content ->
            (   { model
                | character = case content of
                    Ok character -> character
                    Err _ -> model.character
                }
            , Cmd.none
            )
        LinkClicked urlRequest ->
            case urlRequest of
                Browser.Internal url ->
                    ( model
                    , Nav.pushUrl model.key ( Url.toString url ) 
                    )
                Browser.External href ->
                    ( model
                    , Nav.load href 
                    )
        UrlChanged url ->
            (   { model 
                | url = url
                }
            , Cmd.none
            )
        ResetCharacter ->
            (   { model
                | character = Character.new 
                }
            , Cmd.none
            )
        SetDice dice ->
            (   { model
                | dice = dice
                }
            , Cmd.none
            )
        Roll ->
            ( model
            , roll model.dice
            )
        RollResult result ->
            (   { model
                | roll = List.reverse <| List.sort <| result
                }
            , Cmd.none
            )
        ResetRoll ->
            (   { model
                | roll = []
                , dice = Dice.new
                }
            , Cmd.none
            )


roll : Dice -> Cmd Msg
roll dice =
    Random.generate RollResult ( Random.list dice.amount ( Random.int 1 10 ) )

getMode : Url.Url -> Mode
getMode url =
    case url.fragment of
        Just fragment ->
            case fragment of
                "view" -> ViewCharacter
                "create" -> CreateCharacter
                "dice" -> RollDice
                _ -> CreateCharacter
        Nothing ->
            CreateCharacter

getFragment : Url.Url -> String
getFragment url = 
    ( "#" ++ ( Maybe.withDefault "" url.fragment ) )

view : Model -> Browser.Document Msg
view model =
    { title = "Character Sheet"
    , body =
        [ div []
            [ viewNavbar model
            , case getMode model.url of
                ViewCharacter -> viewView model
                CreateCharacter -> viewCreate model
                RollDice -> viewDice model
            ]
        ]
    }

viewNavbar : Model -> Html Msg
viewNavbar model =
    nav 
        [ class "navbar is-fixed-top" 
        , ariaRole "navigation"
        , ariaLabel "main navigation"
        ]
        [ div [ class "navbar-brand" ]
            [ p [ class "navbar-item" ] [ text "MtA Character Creator" ]
            , a 
                [ ariaRole "button"
                , class "navbar-burger"
                , ariaLabel "menu"
                , ariaExpanded "false"
                , href ( getFragment model.url )
                ]
                [ span [ ariaHidden "true" ] []
                , span [ ariaHidden "true" ] []
                , span [ ariaHidden "true" ] []
                , span [ ariaHidden "true" ] []
                ]
            ]
        , div 
            [ id "navbar" 
            , class "navbar-menu"
            ]
            [ div [ class "navbar-start" ]
                [ a 
                    [ class "navbar-item"
                    , href "#create"
                    , onClick ResetCharacter
                    ] 
                    [ text "New Character" ]
                , viewImport model.url
                ]
            , case getMode model.url of
                CreateCharacter -> viewNavbarFinish
                ViewCharacter -> viewNavbarDice
                RollDice -> viewNavbarDiceToChar
            ]
        ]

viewNavbarFinish : Html Msg
viewNavbarFinish =
    div [ class "navbar-end" ]
        [ div [ class "navbar-item" ] 
            [ a 
                [ class "button is-primary" 
                , href "#view"
                ] 
                [ text "Finish Character" ]
            ]
        ]

viewNavbarDice : Html Msg
viewNavbarDice =
    div [ class "navbar-end" ]
        [ div [ class "navbar-item" ] 
            [ a 
                [ class "button" 
                , href "#dice"
                , onClick ResetRoll
                ] 
                [ text "Roll Dice" ]
            ]
        ]

viewNavbarDiceToChar : Html Msg
viewNavbarDiceToChar =
    div [ class "navbar-end" ]
        [ div [ class "navbar-item" ] 
            [ a 
                [ class "button" 
                , href "#view"
                ] 
                [ text "Back to Character" ]
            ]
        ]

viewImport : Url.Url -> Html Msg
viewImport url =
    div [ class "navbar-item has-dropdown is-hoverable" ]
        [ a 
            [ class "navbar-link" 
            , href ( getFragment url )
            ] [ text "Import Character" ]
        , div [ class "navbar-dropdown" ]
            [ a 
                [ class "navbar-item"
                , onClick ( ImportCharacter "./characters/empty.json" )
                , href ( getFragment url )
                ] [ text "Empty" ]
            , a 
                [ class "navbar-item"
                , onClick ( ImportCharacter "./characters/john_doe.json" )
                , href ( getFragment url )
                ] [ text "John Doe" ]
            , a 
                [ class "navbar-item"
                , onClick ( ImportCharacter "./characters/jane_doe.json" )
                , href ( getFragment url )
                ] [ text "Jane Doe" ]
            ]
        ]

viewCreate : Model -> Html Msg
viewCreate model =
    let
        dotSize = 24
    in
        div 
            [ class "container" ] 
            [ viewEditableText Name model.character.name
            , viewAttributes model.character.creationPoints.attributes model.character.stats dotSize
            , viewAbilities model.character.creationPoints.abilities model.character.stats dotSize
            , viewBackgrounds model.url model.activeDropdown model.character.creationPoints ( Stats.backgroundsInList model.character.stats ) dotSize
            , viewSpheres model.character.creationPoints ( Stats.spheresInList model.character.stats ) dotSize
            , viewModal model
            ]

viewEditableText : ModalType -> String -> Html Msg
viewEditableText modalType content =
    div []
        [ span [ class "icon-text" ]
            [ span [] [ text content ]
            , span 
                [ class "icon" 
                ] 
                [ i 
                    [ class "fas fa-pen"
                    , onClick ( OpenModal modalType ) 
                    ] []
                ]
            ]
        ]

viewAttributes : OCat -> List Stat -> Float -> Html Msg
viewAttributes oCat attributes dotSize =
    div [ class "has-text-centered" ]
        [ p [] [ text "Attributes" ]
        , div [ class "columns is-centered is-mobile" ]
            [ viewAttributeCol oCat "Physical" Stats.Physical attributes dotSize
            , viewAttributeCol oCat "Social" Stats.Social attributes dotSize
            , viewAttributeCol oCat "Mental" Stats.Mental attributes dotSize
            ]
        ]

viewAttributeCol : OCat -> String -> Category -> List Stat -> Float -> Html Msg
viewAttributeCol oCat name category attributes dotSize =
    let 
        categoryAttributes : List Stat
        categoryAttributes = Stats.attributesOfCategoryInList category attributes

        index : Int
        index = OCat.indexOf oCat.order category
        basePoints : Int
        basePoints = OCat.categoryValue oCat category
        points : Int
        points = basePoints - ( Stats.count categoryAttributes ) + ( List.length categoryAttributes )
    in
        div [ class "column is-one-fifth" ]
            (   [ p [] [ text name ]
                , creationInfo points
                , div [ class "field has-addons is-fullwidth" ]
                    [ p [ class "control is-expanded" ]
                        [ button 
                            [ class ( "button is-small is-fullwidth" ++ if index == 0 then " is-primary" else "" )
                            , onClick ( ChangePriority OCat.Attribute category 0 )
                            ] [ text "I" ]
                        ]
                    ,  p [ class "control is-expanded" ]
                        [ button 
                            [ class ( "button is-small is-fullwidth" ++ if index == 1 then " is-primary" else "" )
                            , onClick ( ChangePriority OCat.Attribute category 1 )
                            ] [ text "II" ]
                        ]
                    ,  p [ class "control is-expanded" ]
                        [ button 
                            [ class ( "button is-small is-fullwidth" ++ if index == 2 then " is-primary" else "" )
                            , onClick ( ChangePriority OCat.Attribute category 2 )
                            ] [ text "III" ]
                        ]
                    ]
                ]
                ++ 
                ( List.map ( viewStat dotSize ) categoryAttributes
                )
            )

viewAbilities : OCat -> List Stat -> Float -> Html Msg
viewAbilities oCat abilities dotSize =
    div [ class "has-text-centered" ]
        [ p [] [ text "Abilities" ]
        , div [ class "columns is-centered is-desktop" ]
            [ viewAbilityCol oCat "Talents" Stats.Talent abilities dotSize
            , viewAbilityCol oCat "Skills" Stats.Skill abilities dotSize
            , viewAbilityCol oCat "Knowledges" Stats.Knowledge abilities dotSize
            ]
        ]

viewAbilityCol : OCat -> String -> Category -> List Stat -> Float -> Html Msg
viewAbilityCol oCat name category abilities dotSize =
    let 
        categoryAbilities : List Stat
        categoryAbilities = Stats.abilitiesOfCategoryInList category abilities

        index : Int
        index = OCat.indexOf oCat.order category
        basePoints : Int
        basePoints = OCat.categoryValue oCat category
        points : Int
        points = basePoints - ( Stats.count categoryAbilities )
    in
        div [ class "column" ]
            (   [ p [] [ text name ]
                , creationInfo points
                , div [ class "field has-addons is-desktop" ]
                    [ p [ class "control is-expanded" ]
                        [ button 
                            [ class ( "button is-small is-fullwidth" ++ if index == 0 then " is-primary" else "" )
                            , onClick ( ChangePriority OCat.Ability category 0 )
                            ] [ text "I" ]
                        ]
                    ,  p [ class "control is-expanded" ]
                        [ button 
                            [ class ( "button is-small is-fullwidth" ++ if index == 1 then " is-primary" else "" )
                            , onClick ( ChangePriority OCat.Ability category 1 )
                            ] [ text "II" ]
                        ]
                    ,  p [ class "control is-expanded" ]
                        [ button 
                            [ class ( "button is-small is-fullwidth" ++ if index == 2 then " is-primary" else "" )
                            , onClick ( ChangePriority OCat.Ability category 2 )
                            ] [ text "III" ]
                        ]
                    ]
                , div [ class "fixed-grid has-6-cols" ]
                    [ div [ class "grid" ]
                        ( List.map ( viewStatInGrid dotSize 3 ( List.length categoryAbilities ) ) ( List.indexedMap Tuple.pair categoryAbilities )
                        )
                    ]
                ]
            )

viewStatInGrid : Float -> Int -> Int -> ( Int, Stat ) -> Html Msg
viewStatInGrid dotSize cols totalCount indexedStat =
    let
        index : Int
        index = Tuple.first indexedStat
        stat : Stat
        stat = Tuple.second indexedStat

        offset : Int
        offset = Basics.remainderBy cols totalCount
        dif : Int
        dif = totalCount - index

        offsetString : String
        offsetString = 
            if dif == offset
            then " is-col-start-" ++ ( String.fromInt ( 1 + cols - dif ) )
            else ""
    in
        div [ class ( "cell is-col-span-2" ++ offsetString ) ]
            [ div [ class "notification" ]
                [ p [ class "block" ] [ text stat.name ]
                , p [ class "block" ] [ viewDots True dotSize stat ]
                ]
            ]

viewBackgrounds : Url.Url -> Maybe ActiveDropdown -> CreationPoints -> List Stat -> Float -> Html Msg
viewBackgrounds url activeDropdown cp backgrounds dotSize =
    let
        basePoints : Int
        basePoints = cp.backgrounds
        points : Int
        points = basePoints - ( Stats.count backgrounds )

        separatedBackgrounds : ( List Stat, List Stat )
        separatedBackgrounds = Stats.splitZero backgrounds
    in
        div [ class "has-text-centered" ]
            (   [ p [] [ text "Backgrounds" ]
                , div [ class "columns is-centered is-mobile" ]
                    [ div [ class "column is-two-fifths" ]
                        (   [ creationInfo points
                            , viewOtherBackgrounds url activeDropdown ( Tuple.second separatedBackgrounds )
                            ]
                            ++
                            ( List.map ( viewStatHorizontal dotSize ) ( Tuple.first separatedBackgrounds ) )
                        )
                    ]
                ]
            )

viewOtherBackgrounds : Url.Url -> Maybe ActiveDropdown -> List Stat -> Html Msg
viewOtherBackgrounds url activeDropdown backgrounds =
    div [ class ( "block dropdown is-fullwidth" ++ if activeDropdown == Just Backgrounds then " is-active" else "" ) ]
        [ div [ class "dropdown-trigger" ]
            [ div 
                [ class "button" 
                , ariaHaspopup "true"
                , ariaControls "dropdown-menu" 
                , onClick ( SetDropDownActive ( switchDropdownActive activeDropdown Backgrounds ) )
                ]
                [ span [] [ text "Add a Background" ]
                , span [ class "icon is-small" ]
                    [ i 
                        [ class "fas fa-angle-down"
                        , ariaHidden "true"
                        ] [] 
                    ]
                ]
            ]
        , div 
            [ class "dropdown-menu"
            , id "dropdown-menu"
            , ariaRole "menu"
            ]
            [ div [ class "dropdown-content has-text-left" ] 
                ( List.map ( viewStatDropdown url ) backgrounds )
            ]
        ]

viewStatHorizontal : Float -> Stat -> Html Msg
viewStatHorizontal dotSize stat =
    div [ class "notification is-narrow" ]
        [ div [ class "columns is-centered is-vcentered is-mobile" ]
            [ div [ class "column has-text-right" ] [ text stat.name ]
            , div [ class "column has-text-left" ] [ viewDots True dotSize stat ]
            ]
        ]

viewStatDropdown : Url.Url -> Stat -> Html Msg
viewStatDropdown url stat =
    a 
        [ class "dropdown-item" 
        , onClick ( ChangeStat { stat | dots = BoundedInt.changeTo stat.dots 1 } )
        , href ( getFragment url )
        ] [ text stat.name ]
    

viewSpheres : CreationPoints -> List Stat -> Float -> Html Msg
viewSpheres cp spheres dotSize =
    let
        basePoints : Int
        basePoints = cp.spheres
        points : Int
        points = basePoints - ( Stats.count spheres )
    in
        div [ class "has-text-centered" ]
            [ p [] [ text "Spheres" ]
            , div [ class "columns is-centered is-mobile" ]
                [ div [ class "column is-three-fifths" ]
                    [ creationInfo points
                    , affinityWarning spheres
                    ]
                ]
            , div [ class "columns is-centered is-mobile" ]
                [ div [ class "column is-one-fifth" ] 
                    ( List.map ( viewSphere dotSize ) ( Stats.everyThirdStat spheres 0 )
                    )
                , div [ class "column is-one-fifth" ] 
                    ( List.map ( viewSphere dotSize ) ( Stats.everyThirdStat spheres 1 )
                    )
                , div [ class "column is-one-fifth" ] 
                    ( List.map ( viewSphere dotSize ) ( Stats.everyThirdStat spheres 2 )
                    )
                ]
            ]

viewSphere : Float -> Stat -> Html Msg
viewSphere dotSize stat =
    div [ class "notification" ]
        [ p 
            [ class "delete is-not-delete icon corner"
            , onClick ( SetAffinity { stat | statType = Stats.Sphere True } )
            ] 
            [ case stat.statType of
                Stats.Sphere affinity ->
                    if affinity 
                    then i [ class "fas fa-star" ] []
                    else i [ class "far fa-star" ] []
                _ -> i [] []
            ]
        , p [ class "block" ] [ text stat.name ]
        , p [ class "block" ] [ viewDots True dotSize stat ]
        ]

affinityWarning : List Stat -> Html Msg
affinityWarning spheres = 
    let
        hasSetAffinity : Bool
        hasSetAffinity = Stats.hasSetAffinity spheres
    in
        if hasSetAffinity
        then div [] []
        else div [ class "notification is-danger" ] [ text "Affinity Sphere not set" ]

creationInfo : Int -> Html Msg
creationInfo points = 
    let
        notificationColor : String
        notificationColor =
            case compare points 0 of
                GT -> "is-warning"
                EQ -> "is-success"
                LT -> "is-danger"
    in
        div [ class ( "notification " ++ notificationColor ) ]
            [ text ( ( String.fromInt points ) ++ " remaining" )
            ]

viewStat : Float -> Stat -> Html Msg
viewStat dotSize stat =
    div [ class "notification" ]
        [ p [ class "block" ] [ text stat.name ]
        , p [ class "block" ] [ viewDots True dotSize stat ]
        ]

viewDots : Bool -> Float -> Stat -> Html Msg
viewDots allowChange dotSize stat =
    let
        dotGap = dotSize / 2
        dotOutline = 0.05 * dotSize
        marginalIncrease = (  dotSize + dotGap ) / 2
        svgOutlineMargin = ( sqrt ( 2 * ( dotOutline ^ 2 ) ) ) / 2

        rangeList = List.range 0 ( stat.dots.max - 1 )

        isFilledList = rangeList |> List.map ( (>) stat.dots.value )
    in
        svg 
            [ Svg.Attributes.width  ( String.fromFloat ( dotSize + marginalIncrease * toFloat ( stat.dots.max - 1 ) + 2 * svgOutlineMargin ) )
            , Svg.Attributes.height ( String.fromFloat ( dotSize + marginalIncrease + 2 * svgOutlineMargin ) )
            ]
            ( rangeList
                |> List.map ( toDotPosition dotSize dotGap svgOutlineMargin )
                |> List.map3 ( viewDot allowChange stat dotSize dotOutline ) isFilledList rangeList
            )

toDotPosition : Float -> Float -> Float -> Int -> ( Float, Float )
toDotPosition size gap offset id =
    let
        marginalIncrease = ( size + gap ) / 2
    in
        ( toFloat id * marginalIncrease + offset
        , toFloat ( remainderBy 2 ( id + 1 ) ) * marginalIncrease + offset
        )

viewDot : Bool -> Stat -> Float -> Float -> Bool -> Int -> ( Float, Float ) -> Html Msg
viewDot allowChange stat size strokeWidth isFilled id position =
    Svg.polygon 
        [ Svg.Attributes.points 
            (   [ ( 0.5, 0.0 )
                , ( 1.0, 0.5 )
                , ( 0.5, 1.0 )
                , ( 0.0, 0.5 )
                ]
                |> List.map ( Tuple.mapBoth   ( (*) size ) ( (*) size ) )
                |> List.map ( Tuple.mapFirst  ( (+) ( Tuple.first  position ) ) )
                |> List.map ( Tuple.mapSecond ( (+) ( Tuple.second position ) ) )
                |> List.map tupleToString
                |> List.intersperse " "
                |> String.concat
            )
        , Svg.Attributes.fill "var(--bulma-body-color)"
        , Svg.Attributes.fillOpacity ( boolAsIntString isFilled )
        , Svg.Attributes.stroke "var(--bulma-body-color)"
        , Svg.Attributes.strokeWidth ( String.fromFloat strokeWidth )
        , onClick 
            ( ChangeStat 
                ( if allowChange
                    then { stat | dots = BoundedInt.changeTo stat.dots ( valueOrReset stat.dots ( id + 1 ) ) }
                    else stat
                )
            )
        ] []

tupleToString : ( Float, Float ) -> String
tupleToString tuple =
    let
        first  = String.fromFloat ( Tuple.first  tuple )
        second = String.fromFloat ( Tuple.second tuple )
    in
        first ++ "," ++ second

boolAsIntString : Bool -> String
boolAsIntString bool =
    if bool
    then "1"
    else "0"

valueOrReset : BoundedInt -> Int -> Int
valueOrReset dots value =
    if dots.value == value
    then dots.min
    else value

switchDropdownActive : Maybe ActiveDropdown -> ActiveDropdown -> Maybe ActiveDropdown
switchDropdownActive activeDropdown dropdownAction =
    if activeDropdown == Just dropdownAction
    then Nothing
    else Just dropdownAction

viewModal : Model -> Html Msg
viewModal model =
    case model.modal of
        Nothing -> div [ class "modal" ] []
        Just modalType ->
            div [ class "modal is-active" ]
                [ div 
                    [ class "modal-background" 
                    , onClick CloseModal
                    ] []
                , div [ class "modal-card" ] 
                    [ header [ class "modal-card-head" ]
                        [ p [ class "modal-card-title" ] [ text ( "Change " ++ ( modalTitle modalType ) ) ]
                        ]
                    , section [ class "modal-card-body" ] 
                        [ p [] 
                            [ input 
                                [ class "input" 
                                , type_ "text"
                                , placeholder ( "Enter New " ++ ( modalTitle modalType ) )
                                , value model.modalValue
                                , onInput ( \x -> ChangeModalValue x )
                                ] []
                            ]
                        ]
                    , footer [ class "modal-card-foot" ] 
                        [ div [ class "buttons" ] 
                            [ button 
                                [ class "button is-success"
                                , onClick ( SaveModalChange modalType )
                                ] [ text "Save Changes" ]
                            , button 
                                [ class "button"
                                , onClick CloseModal
                                ] [ text "Cancel" ]
                            ]
                        ]
                    ]
                ]

modalTitle : ModalType -> String
modalTitle modalType =
    case modalType of
        Name -> "Character Name"

modalValue : ModalType -> Character -> String
modalValue modalType character =
    case modalType of
        Name -> character.name

viewView : Model -> Html Msg
viewView model =
    let
        dotSize = 16
    in
        div [ class "container" ] 
            [ vViewHeader model.character dotSize
            , vViewBody model.character dotSize
            ]

vViewHeader : Character -> Float -> Html Msg
vViewHeader character dotSize =
    div [ class "block "]
        [ nav [ class "columns is-multiline" ]
            [ div [ class "column" ]
                [ div [] 
                    [ div [ class "level" ] 
                        [ div [ class "level-left" ]
                            [ div [ class "level-item" ]
                                [ div [] 
                                    [ p [] [ text character.name ]
                                    , p [] [ text "Default Chronicle" ]
                                    , p [] [ text "Default Player" ]
                                    ]
                                ]
                            ]
                        , div [ class "level-right" ]
                            [ div [ class "level-item" ]
                                [ div [ class "columns is-multiline is-right" ]
                                    [ div [ class "column is-narrow" ] 
                                        [ div [ class "buttons has-addons" ] 
                                            [ button [ class "button is-static" ] [ text "Downtime" ]
                                            , button [ class "button" ] [ text "0" ]
                                            ] 
                                        ]
                                    , div [ class "column is-narrow" ] 
                                        [ div [ class "buttons has-addons" ] 
                                            [ button [ class "button is-static" ] [ text "Tass" ]
                                            , button [ class "button" ] [ text "0" ]
                                            ] 
                                        ]
                                    , div [ class "column is-narrow" ] 
                                        [ div [ class "buttons has-addons" ] 
                                            [ button [ class "button is-static" ] [ text "Health" ]
                                            , button [ class "button" ] [ text "0" ]
                                            ] 
                                        ]
                                    , div [ class "column is-narrow" ] 
                                        [ div [ class "buttons has-addons" ] 
                                            [ button [ class "button is-static" ] [ text "Experience" ]
                                            , button [ class "button" ] [ text "0" ]
                                            ]
                                        ]
                                    , div [ class "column is-narrow" ] 
                                        [ button [ class "button" ] [ text "Lvl Up" ]
                                        ]
                                    , div [ class "column is-narrow" ] 
                                        [ button [ class "button" ] [ text "Edit" ]
                                        ]
                                    ]
                                ]
                            ]
                        ]
                    , vViewSpheres ( Stats.spheresInList character.stats ) dotSize
                    ]
                ]
            ]
        ]

vViewSpheres : List Stat -> Float -> Html Msg
vViewSpheres spheres dotSize =
    div [ class "columns has-text-centered"] 
        ( List.map ( vViewStatAsCol dotSize ) spheres 
        )

vViewStatAsCol : Float -> Stat -> Html Msg
vViewStatAsCol dotSize stat =
    div [ class "column" ]
        [ div [ class "notification" ]
            [ p [ class "block" ] [ text ( first4Letters stat.name ) ]
            , p [ class "block" ] [ viewDots False dotSize stat ]
            ]
        ]

first4Letters : String -> String
first4Letters string =
    String.slice 0 4 string
    |> String.toUpper

vViewBody : Character -> Float -> Html Msg
vViewBody character dotSize = 
    div [ class "columns" ]
        [ div [ class "column is-half" ] 
            [ div [ class "columns" ]
                [ div [ class "column is-half" ] 
                    [ vViewStatHorizontal ( Stats.areteInList character.stats ) dotSize
                    ]
                , div [ class "column is-half" ] 
                    [ vViewStatHorizontal { name = "Willpower", dots = BoundedInt.new 5 0 10, statType = Stats.Arete } dotSize
                    ]
                ]
            , vViewQuintParadox dotSize
            , div [ class "columns" ]
                [ div [ class "column is-half" ]
                    [ vViewStatList ( Tuple.first <| Stats.splitZero <| Stats.backgroundsInList character.stats ) "Backgrounds" dotSize
                    ]
                , div [ class "column is-half" ] 
                    [ vViewStatList [] "Merits" dotSize
                    , vViewStatList [] "Flaws" dotSize
                    ]
                ]
            , div [ class "columns" ]
                [ div [ class "column is-one-third" ]
                    [ vViewNDE "Nature" "NaN"
                    ]
                , div [ class "column is-one-third" ]
                    [ vViewNDE "Demeanor" "NaN"
                    ]
                , div [ class "column is-one-third" ]
                    [ vViewNDE "Essence" "NaN"
                    ]
                ]
            , vViewNotes ""
            ]
        , div [ class "column is-half" ] 
            [ vViewAttributes ( Stats.attributesInList character.stats ) dotSize
            , vViewAbilities ( Stats.abilitiesInList character.stats ) dotSize
            ]
        ]

vViewStatHorizontal : Stat -> Float -> Html Msg
vViewStatHorizontal stat dotSize =
    div [ class "notification" ]
        [ div [ class "columns is-vcentered has-text-left"]
            [ div [ class "column" ] [ text stat.name ]
            , div [ class "column" ] [ viewDots False dotSize stat ]
            ]
        ]

vViewQuintParadox : Float -> Html Msg
vViewQuintParadox dotSize = 
    div [ class "notification" ] 
        [ nav [ class "level" ] 
            [ div [ class "level-left" ] 
                [ div [ class "level-item" ] [ text "Quintessence" ]
                ]
            , div [ class "level-center is-centered" ]
                [ viewDots False dotSize { name = "", dots = BoundedInt.new 0 0 20, statType = Stats.Arete }
                ]
            , div [ class "level-right" ]
                [ div [ class "level-item" ] [ text "Paradox" ]
                ]
            ]
        ] 

vViewNDE : String -> String -> Html Msg
vViewNDE category value =
    div [ class "notification has-text-centered" ]
        [ div [ class "level-item" ] [ text category ]
        , div [ class "level-item" ] [ text value ]
        ]

vViewNotes : String -> Html Msg
vViewNotes notes = 
    div [ class "notification" ]
        [ div [] [ text "Notes" ]
        , div []
            [ textarea 
                [ class "textarea"
                , placeholder "Character Notes"
                ] 
                [ text notes ]
            ]
        ]

vViewStatList : List Stat -> String -> Float -> Html Msg
vViewStatList stats statTypeName dotSize =
    div [ class "notification" ]
        [ table [ class "table is-fullwidth" ]
            [ thead []
                [ tr []
                    [ th [ colspan 2 ] [ text statTypeName ]
                    ]
                ]
            , tbody []
                ( List.map ( vViewStat dotSize ) stats )
            ]
        ]

vViewStat : Float -> Stat -> Html Msg
vViewStat dotSize stat =
    tr []
        [ td [] [ text stat.name ]
        , td [] [ viewDots False dotSize stat ]
        ]

vViewAttributes : List Stat -> Float -> Html Msg
vViewAttributes attributes dotSize =
    div [ class "notification" ]
        [ table [ class "table is-fullwidth" ]
            [ thead []
                [ tr []
                    [ th [ colspan 2, style "width" "33%" ] [ text "Physical" ]
                    , th [ colspan 2, style "width" "33%" ] [ text "Social" ]
                    , th [ colspan 2, style "width" "33%" ] [ text "Mental" ]
                    ]
                ]
            , tbody []
                ( List.map3 ( vView3ColTableStat dotSize )
                    ( Stats.attributesOfCategoryInList Stats.Physical attributes )
                    ( Stats.attributesOfCategoryInList Stats.Social attributes )
                    ( Stats.attributesOfCategoryInList Stats.Mental attributes )
                )
            ]
        ]

vView3ColTableStat : Float -> Stat -> Stat -> Stat -> Html Msg
vView3ColTableStat dotSize a b c =
    tr []
        [ td [] [ text a.name ]
        , td [] [ viewDots False dotSize a ]
        , td [] [ text b.name ]
        , td [] [ viewDots False dotSize b ]
        , td [] [ text c.name ]
        , td [] [ viewDots False dotSize c ]
        ]

vViewAbilities : List Stat -> Float -> Html Msg
vViewAbilities abilities dotSize =
    div [ class "notification" ]
        [ table [ class "table is-fullwidth" ]
            [ thead []
                [ tr []
                    [ th [ colspan 2, style "width" "33%" ] [ text "Talents" ]
                    , th [ colspan 2, style "width" "33%" ] [ text "Skills" ]
                    , th [ colspan 2, style "width" "33%" ] [ text "Knowledges" ]
                    ]
                ]
            , tbody []
                ( List.map3 ( vView3ColTableStat dotSize )
                    ( Stats.abilitiesOfCategoryInList Stats.Talent abilities )
                    ( Stats.abilitiesOfCategoryInList Stats.Skill abilities )
                    ( Stats.abilitiesOfCategoryInList Stats.Knowledge abilities )
                )
            ]
        ]

viewDice : Model -> Html Msg
viewDice model =
    div [ class "container" ] 
        [ div [ class "columns is-centered" ]
            [ div [ class "column is-one-third" ]
                [ viewDiceInput model
                ]
            , div [ class "column is-two-thirds" ]
                [ viewDiceRoll model.dice model.roll
                ]
            ]
        ]

viewDiceInput : Model -> Html Msg
viewDiceInput model = 
    div [] 
        [ div [ class "field is-horizontal" ] 
            [ div [ class "field-label is-normal" ]
                [ div [ class "label" ] [ text "Amount" ] ]
            , div [ class "field-body" ]
                [ div [ class "field" ]
                    [ div [ class "control" ] 
                        [ input 
                            [ class "input"
                            , type_ "number"
                            , Html.Attributes.min "1"
                            , placeholder "1"
                            , onInput ( \x -> SetDice ( Dice.changeAmountTo model.dice ( Maybe.withDefault 1 <| String.toInt x ) ) )
                            ] []
                        ]
                    ]
                ]
            ]
        , div [ class "field is-horizontal" ] 
            [ div [ class "field-label is-normal" ]
                [ div [ class "label" ] [ text "Difficulty" ] ]
            , div [ class "field-body" ]
                [ div [ class "field" ]
                    [ div [ class "control" ] 
                        [ input 
                            [ class "input"
                            , type_ "number"
                            , Html.Attributes.min "1"
                            , Html.Attributes.max "10"
                            , placeholder "6"
                            , onInput ( \x -> SetDice ( Dice.changeDifficultyTo model.dice ( Maybe.withDefault 6 <| String.toInt x ) ) )
                            ] []
                        ]
                    ]
                ]
            ]
        , div [ class "field" ] 
            [ label [ class "checkbox" ]
                [ input 
                    [ type_ "checkbox" 
                    , onClick ( SetDice ( Dice.switchWillpower model.dice ) )
                    ] []
                , text "Use Willpower"
                ]
            ]
        , div [ class "field" ] 
            [ label [ class "checkbox" ]
                [ input 
                    [ type_ "checkbox" 
                    , onClick ( SetDice ( Dice.switchSpecialty model.dice ) )
                    ] []
                , text "Is Specialty"
                ]
            ]
        , div [ class "field is-horizontal" ] 
            [ div [ class "field-body" ]
                [ div [ class "field" ]
                    [ div [ class "control" ] 
                        [ button 
                            [ class "button is-primary is-fullwidth"
                            , onClick Roll
                            ] 
                            [ text "Roll Dice"
                            ]
                        ]
                    ]
                ]
            ]
        ]

viewDiceRoll : Dice -> List Int -> Html Msg
viewDiceRoll dice rollResult =
    if List.length rollResult > 0
    then
        div [] 
            [ viewSuccesses ( Dice.successes dice rollResult )
            , div [ class "square-container" ] 
                ( List.map ( viewResultDie dice.difficulty ) rollResult )
            ]
    else
        div [] []

viewSuccesses : Int -> Html Msg
viewSuccesses successes =
    let
        color : String
        color =
            if successes > 0
            then " is-success"
            else " is-danger"
        successText : String
        successText =
            if successes == 0
            then "Failure"
            else
                if successes > 0
                then
                    if successes > 1
                    then ( String.fromInt successes ) ++ " Successes"
                    else "1 Success"
                else
                    if successes < -1
                        then ( String.fromInt <| negate successes ) ++ " Botches"
                        else "1 Botch"
    in
        div 
            [ class ( "notification" ++ color ) ] 
            [ text successText ]

viewResultDie : Int -> Int -> Html Msg
viewResultDie difficulty result =
    let
        color : String
        color =
            if result >= difficulty
            then " is-success"
            else
                if result == 1
                then " is-danger"
                else ""
    in
        div 
            [ class "square" ] 
            [ div [ class ( "content notification" ++ color ) ]
                [ div [ class "label is-large" ] [ text ( String.fromInt result ) ]
                ]
            ]

-- subscriptions

subscriptions : Model -> Sub Msg
subscriptions _ =
  Sub.none

-- aria functions

ariaLabel : String -> Attribute msg
ariaLabel value =
    Html.Attributes.attribute "aria-label" value

ariaHaspopup : String -> Attribute msg
ariaHaspopup value =
    Html.Attributes.attribute "aria-haspopup" value

ariaControls : String -> Attribute msg
ariaControls value =
    Html.Attributes.attribute "aria-controls" value

ariaHidden : String -> Attribute msg
ariaHidden value =
    Html.Attributes.attribute "aria-hidden" value

ariaRole : String -> Attribute msg
ariaRole value =
    Html.Attributes.attribute "role" value

ariaExpanded : String -> Attribute msg
ariaExpanded value =
    Html.Attributes.attribute "aria-expanded" value

-- json decoder

jsonDecoder : Json.Decode.Decoder Character
jsonDecoder =
    Json.Decode.map3 Character
        ( Json.Decode.field   "name"  Json.Decode.string )
        ( Json.Decode.succeed         Creation.new       )
        ( Json.Decode.field   "stats" jsonStatsDecoder   )


jsonStatsDecoder : Json.Decode.Decoder ( List Stat )
jsonStatsDecoder =
    Json.Decode.map5 ( \spheres arete attributes abilities backgrounds -> spheres ++ ( List.singleton arete ) ++ attributes ++ abilities ++ backgrounds )
        ( Json.Decode.field "spheres"     <| Json.Decode.list jsonSphereDecoder     )
        ( Json.Decode.field "arete"          jsonAreteDecoder                       )
        ( Json.Decode.field "attributes"  <| jsonAttributesDecoder                  )
        ( Json.Decode.field "abilities"   <| jsonAbilitiesDecoder                   )
        ( Json.Decode.field "backgrounds" <| Json.Decode.list jsonBackgroundDecoder )

jsonSphereDecoder : Json.Decode.Decoder Stat
jsonSphereDecoder =
    Json.Decode.map3 ( \name dots affinity -> Stats.newSphere name |> Stats.changeDotsTo dots |> Stats.setAffinityNoChecks affinity )
        ( Json.Decode.field "name"     Json.Decode.string )
        ( Json.Decode.field "dots"     Json.Decode.int    )
        ( Json.Decode.field "affinity" Json.Decode.bool   )

jsonAreteDecoder : Json.Decode.Decoder Stat
jsonAreteDecoder =
    Json.Decode.map ( \dots -> Stats.newArete |> Stats.changeDotsTo dots )
        ( Json.Decode.field "dots" Json.Decode.int )

jsonAttributesDecoder : Json.Decode.Decoder ( List Stat )
jsonAttributesDecoder =
    Json.Decode.map3 ( \physical social mental -> physical ++ social ++ mental )
        ( Json.Decode.field "physical" <| Json.Decode.list ( jsonAttributeDecoder Stats.Physical ) )
        ( Json.Decode.field "social"   <| Json.Decode.list ( jsonAttributeDecoder Stats.Social   ) )
        ( Json.Decode.field "mental"   <| Json.Decode.list ( jsonAttributeDecoder Stats.Mental   ) )

jsonAttributeDecoder : Stats.Category -> Json.Decode.Decoder Stat
jsonAttributeDecoder cat =
    Json.Decode.map2 ( \name dots -> Stats.newAttribute cat name |> Stats.changeDotsTo dots )
        ( Json.Decode.field "name" Json.Decode.string )
        ( Json.Decode.field "dots" Json.Decode.int    )

jsonAbilitiesDecoder : Json.Decode.Decoder ( List Stat )
jsonAbilitiesDecoder =
    Json.Decode.map3 ( \talents skills knowledges -> talents ++ skills ++ knowledges )
        ( Json.Decode.field "talents"    <| Json.Decode.list ( jsonAbilityDecoder Stats.Talent    ) )
        ( Json.Decode.field "skills"     <| Json.Decode.list ( jsonAbilityDecoder Stats.Skill     ) )
        ( Json.Decode.field "knowledges" <| Json.Decode.list ( jsonAbilityDecoder Stats.Knowledge ) )

jsonAbilityDecoder : Stats.Category -> Json.Decode.Decoder Stat
jsonAbilityDecoder cat =
    Json.Decode.map2 ( \name dots -> Stats.newAbility cat name |> Stats.changeDotsTo dots )
        ( Json.Decode.field "name" Json.Decode.string )
        ( Json.Decode.field "dots" Json.Decode.int    )

jsonBackgroundDecoder : Json.Decode.Decoder Stat
jsonBackgroundDecoder =
    Json.Decode.map3 ( \name dots max -> Stats.newBackground name max |> Stats.changeDotsTo dots )
        ( Json.Decode.field "name" Json.Decode.string )
        ( Json.Decode.field "dots" Json.Decode.int    )
        ( Json.Decode.field "max"  Json.Decode.int    )